# test_workers.py

import pytest

import time
from curio import *
import pytest
import sys

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n - 1) + fib(n - 2)

#@pytest.mark.skipif(sys.platform.startswith("win"),
#                    reason='broken on Windows')
def test_cpu(kernel):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def cpu_bound(n):
        r = await run_in_process(fib, n)
        results.append(('fib', r))

    async def main():
        async with TaskGroup() as g:
            await g.spawn(spin, 10)
            await g.spawn(cpu_bound, 36)

    kernel.run(main())

    assert results == [
        10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
        ('fib', 14930352)
    ]

#@pytest.mark.skipif(sys.platform.startswith("win"),
#                    reason='broken on Windows')
def test_bad_cpu(kernel):
    async def main():
        with pytest.raises(TypeError):
            r = await run_in_process(fib, '1')

    kernel.run(main())


def test_blocking(kernel):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def blocking(n):
        await run_in_thread(time.sleep, n)
        results.append('sleep done')

    async def main():
        async with TaskGroup() as g:
            await g.spawn(spin, 10)
            await g.spawn(blocking, 2)

    kernel.run(main())

    assert results == [
        10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
        'sleep done',
    ]

def test_executor(kernel):
    from concurrent.futures import ThreadPoolExecutor
    pool = ThreadPoolExecutor()
    async def main():
        r = await run_in_executor(pool, fib, 1)
        assert r == 1
    kernel.run(main())

@pytest.mark.parametrize('runner', [run_in_thread, run_in_process ])
def test_worker_cancel(kernel, runner):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def blocking(n):
        task = await spawn(runner, time.sleep, n)
        await sleep(0.55)
        await task.cancel()
        try:
            await task.join()
        except TaskError as e:
            if isinstance(e.__cause__, CancelledError):
                results.append('cancel')
            else:
                results.append(repr(e.__cause__))

    async def main():
        async with TaskGroup() as g:
            await g.spawn(spin, 10)
            await g.spawn(blocking, 5)

    kernel.run(main())

    assert results == [
        10, 9, 8, 7, 6, 5, 'cancel', 4, 3, 2, 1
    ]


@pytest.mark.parametrize('runner', [run_in_thread, run_in_process])
def test_worker_timeout(kernel, runner):
    results = []

    async def spin(n):
        while n > 0:
            results.append(n)
            await sleep(0.1)
            n -= 1

    async def blocking(n):
        try:
            result = await timeout_after(0.55, runner(time.sleep, n))
        except TaskTimeout:
            results.append('cancel')

    async def main():
        async with TaskGroup() as g:
            await g.spawn(spin, 10)
            await g.spawn(blocking, 5)

    kernel.run(main())

    assert results == [
        10, 9, 8, 7, 6, 5, 'cancel', 4, 3, 2, 1
    ]


def test_exception(kernel):
    results = []

    async def error():
        try:
            result = await run_in_thread(fib, '10')
            results.append('fail')
        except Exception as e:
            results.append(type(e))
            results.append(e.__cause__)

    async def main():
        await error()

    kernel.run(main())

    assert results == [
        TypeError,
        None
    ]



# test_kernel.py
import sys
import time
import pytest
from curio import *
import sys


def test_cancel_noblock(kernel):
    cancelled = False
    child_exit = False

    async def child():
        nonlocal child_exit
        try:
            await sleep(10)
        finally:
            assert cancelled
            child_exit = True

    async def main():
        nonlocal cancelled
        t = await spawn(child)
        await t.cancel(blocking=False)
        cancelled = True
        await t.cancel()

    kernel.run(main)
    assert child_exit


def test_cancel_custom_exception(kernel):
    cancelled = False
    child_exit = False

    class MyException(Exception):
        pass

    async def child():
        nonlocal child_exit
        try:
            await sleep(10)
        except MyException:
            child_exit = True
            return
        child_exit = False

    async def main():
        t = await spawn(child)
        await t.cancel(exc=MyException)

    kernel.run(main)
    assert child_exit

def test_cancel_custom_exception_instance(kernel):
    cancelled = False
    child_exit = False

    class MyException(Exception):
        pass

    exc = MyException("Test")

    async def child():
        nonlocal child_exit
        try:
            await sleep(10)
        except MyException as e:
            child_exit = e is exc
            return
        child_exit = False

    async def main():
        t = await spawn(child)
        await t.cancel(exc=exc)

    kernel.run(main)
    assert child_exit

def test_task_group(kernel):
    async def child(x, y):
        return x + y

    async def main():
        async with TaskGroup() as g:
            t1 = await g.spawn(child, 1, 1)
            t2 = await g.spawn(child, 2, 2)
            t3 = await g.spawn(child, 3, 3)

        assert t1.result == 2
        assert t2.result == 4
        assert t3.result == 6
        assert g.results == [2, 4, 6]


    kernel.run(main())


def test_task_group_daemon(kernel):
    async def child(x, y):
        return x + y

    async def main():
        async with TaskGroup(wait=all) as g:
            t1 = await g.spawn(child, 1, 1)
            t2 = await g.spawn(child, 2, 2, daemon=True)
            t3 = await g.spawn(child, 3, "3", daemon=True)
            t4 = await g.spawn(sleep, 0.5, daemon=True)

        assert t1.result == 2
        assert g.results == [2]


    kernel.run(main())

def test_task_group_existing(kernel):
    evt = Event()
    async def child(x, y):
        return x + y

    async def child2(x, y):
        await evt.wait()
        return x + y

    async def main():
        t1 = await spawn(child, 1, 1)
        t2 = await spawn(child2, 2, 2)
        t3 = await spawn(child2, 3, 3)
        t4 = await spawn(child, 4, 4)
        await t1.join()
        await t4.join()

        async with TaskGroup([t1, t2, t3]) as g:
            await evt.set()
            await g.add_task(t4)

        assert t1.result == 2
        assert t2.result == 4
        assert t3.result == 6
        assert t4.result == 8
        assert g.results == [2,4,6,8]

    kernel.run(main())

def test_task_any_cancel(kernel):
    evt = Event()
    async def child(x, y):
        return x + y

    async def child2(x, y):
        await evt.wait()
        return x + y

    async def main():
        async with TaskGroup(wait=any) as g:
            t1 = await g.spawn(child, 1, 1)
            t2 = await g.spawn(child2, 2, 2)
            t3 = await g.spawn(child2, 3, 3)

        assert t1.result == 2
        assert t1 == g.completed
        assert t2.cancelled
        assert t3.cancelled

    kernel.run(main())


def test_task_any_error(kernel):
    evt = Event()
    async def child(x, y):
        return x + y

    async def child2(x, y):
        await evt.wait()
        return x + y

    async def main():
        async with TaskGroup(wait=any) as g:
            t1 = await g.spawn(child, 1, '1')
            t2 = await g.spawn(child2, 2, 2)
            t3 = await g.spawn(child2, 3, 3)
        try:
            result = g.result
            assert False
        except TypeError:
            assert True
        else:
            assert False

        assert isinstance(g.exception, TypeError)
        assert t1.exception
        assert t2.cancelled
        assert t3.cancelled
        assert all(g.exceptions)

    kernel.run(main())


def test_task_group_iter(kernel):
    async def child(x, y):
        return x + y

    async def main():
        results = set()
        async with TaskGroup() as g:
            await g.spawn(child, 1, 1)
            await g.spawn(child, 2, 2)
            await g.spawn(child, 3, 3)
            async for task in g:
                results.add(task.result)

        assert results == { 2, 4, 6 }
        assert g.results == []    # Explicit collection of results prevents collections on the group

    kernel.run(main())


def test_task_wait_none(kernel):
    evt = Event()
    async def child2(x, y):
        await evt.wait()
        return x + y

    async def main():
        async with TaskGroup(wait=None) as g:
            t2 = await g.spawn(child2, 2, 2)
            t3 = await g.spawn(child2, 3, 3)
        assert t2.cancelled
        assert t3.cancelled

    kernel.run(main())

def test_task_join_daemon(kernel):
    async def child(x, y):
        await sleep(0.1)
        return x + y

    async def main():
        async with TaskGroup(wait=all) as g:
            t2 = await g.spawn(child, 2, 2, daemon=True)
            t3 = await g.spawn(child, 3, 3)
            r = await t2.join()
        assert g.results == [6]

    kernel.run(main())


def test_task_group_error(kernel):
    evt = Event()
    async def child(x, y):
        result = x + y
        await evt.wait()

    async def main():
        async with TaskGroup() as g:
            t1 = await g.spawn(child, 1, 1)
            t2 = await g.spawn(child, 2, 2)
            t3 = await g.spawn(child, 3, 'bad')

        assert isinstance(t3.exception, TypeError)
        assert g.completed == t3
        assert t1.cancelled
        assert t2.cancelled

    kernel.run(main())


def test_task_group_error_block(kernel):
    evt = Event()
    async def child(x, y):
        result = x + y
        await evt.wait()

    async def main():
        try:
            async with TaskGroup() as g:
                t1 = await g.spawn(child, 1, 1)
                t2 = await g.spawn(child, 2, 2)
                t3 = await g.spawn(child, 3, 3)
                raise RuntimeError()
        except RuntimeError:
            assert True
        else:
            assert False
        assert t1.cancelled
        assert t2.cancelled
        assert t3.cancelled

    kernel.run(main())

def test_task_group_join(kernel):
    evt = Event()
    async def child(x, y):
        result = x + y
        await evt.wait()
        return result

    async def main():
        async with TaskGroup() as g:
            t1 = await g.spawn(child, 1, 'foo')
            t2 = await g.spawn(child, 2, 2)
            t3 = await g.spawn(child, 3, 3)
            try:
                await t1.join()
            except TaskError as e:
                assert isinstance(e.__cause__, TypeError)
                assert isinstance(t1.exception, TypeError)
                with pytest.raises(TypeError):
                    t1.result
            else:
                assert False

            # These assert that the error has not cancelled other tasks
            with pytest.raises(RuntimeError):
                t2.result

            with pytest.raises(RuntimeError):
                t2.exception

            await evt.set()

        # Assert that other tasks ran to completion
        assert not t2.cancelled
        assert not t3.cancelled
        assert g.results == [4, 6]


    kernel.run(main())

def test_task_group_cancel(kernel):
    evt = Event()
    evt2 = Event()
    async def child():
        try:
            await evt.wait()
        except CancelledError:
            assert True
            raise
        else:
            raise False

    async def coro():
        try:
            async with TaskGroup() as g:
                t1 = await g.spawn(child)
                t2 = await g.spawn(child)
                t3 = await g.spawn(child)
                await evt2.set()
        except CancelledError:
            assert t1.cancelled
            assert t2.cancelled
            assert t3.cancelled
            raise
        else:
            assert False

    async def main():
        t = await spawn(coro)
        await evt2.wait()
        await t.cancel()

    kernel.run(main)


def test_task_group_timeout(kernel):
    evt = Event()
    async def child():
        try:
            await evt.wait()
        except TaskCancelled:
            assert True
            raise
        else:
            raise False

    async def coro():
        try:
            async with timeout_after(0.25):
                try:
                    async with TaskGroup() as g:
                        t1 = await g.spawn(child)
                        t2 = await g.spawn(child)
                        t3 = await g.spawn(child)
                except CancelledError:
                    assert t1.cancelled
                    assert t2.cancelled
                    assert t3.cancelled
                    raise
        except TaskTimeout:
            assert True
        else:
            assert False

    kernel.run(coro)


def test_task_group_cancel_remaining(kernel):
    evt = Event()
    async def child(x, y):
        return x + y

    async def waiter():
        await evt.wait()

    async def main():
        async with TaskGroup() as g:
            t0 = await g.spawn(child, 1, 1)
            t1 = await g.spawn(child, 2, 2)
            t2 = await g.spawn(waiter)
            t3 = await g.spawn(waiter)
            t = await g.next_done()
            assert t == t0
            r = await g.next_result()
            assert r == 4
            await g.cancel_remaining()

        assert t2.cancelled
        assert t3.cancelled

    kernel.run(main)

def test_task_group_use_error(kernel):
    async def main():
         async with TaskGroup() as g:
              t1 = await g.spawn(sleep, 0)
              with pytest.raises(RuntimeError):
                  await g.add_task(t1)

         with pytest.raises(RuntimeError):
             await g.spawn(sleep, 0)

         t2 = await spawn(sleep, 0)
         with pytest.raises(RuntimeError):
             await g.add_task(t2)
         await t2.join()

         with pytest.raises(RuntimeError):
             await g.spawn_thread(time.sleep, 0)

    kernel.run(main())

def test_task_group_empty(kernel):
    async def main():
        async with TaskGroup() as g:
            pass

        assert g.exception is None
        assert g.exceptions == []
        assert g.results == []
        with pytest.raises(RuntimeError):
             g.result

    kernel.run(main())

def test_defer_cancellation(kernel):
    async def cancel_me(e1, e2):
        with pytest.raises(CancelledError):
            async with disable_cancellation():
                await e1.set()
                await e2.wait()
            await check_cancellation()

    async def main():
        e1 = Event()
        e2 = Event()
        task = await spawn(cancel_me, e1, e2)
        await e1.wait()
        await task.cancel(blocking=False)
        await e2.set()
        await task.join()

    kernel.run(main)

def test_explicit_raise_cancel_disable(kernel):
    async def cancel_me(e1, e2):
        with pytest.raises(CancelledError):
            async with disable_cancellation():
                async with disable_cancellation():
                    await e1.set()
                    await e2.wait()
                    raise TaskCancelled()
                # Even if TaskCancelled is raised explicitly, it should not propagate out.
                # Instead, it gets defer until cancellation is enabled
                assert await check_cancellation()
                assert True
            await check_cancellation()

    async def main():
        e1 = Event()
        e2 = Event()
        task = await spawn(cancel_me, e1, e2)
        await e1.wait()
        await task.cancel(blocking=False)
        await e2.set()
        await task.join()

    kernel.run(main)

def test_disable_cancellation_function(kernel):
    async def cancel_it(e1, e2):
        await e1.set()
        await e2.wait()

    async def cancel_me(e1, e2):
        with pytest.raises(CancelledError):
            await disable_cancellation(cancel_it(e1, e2))
            await check_cancellation()

    async def main():
        e1 = Event()
        e2 = Event()
        task = await spawn(cancel_me, e1, e2)
        await e1.wait()
        await task.cancel(blocking=False)
        await e2.set()
        await task.join()

    kernel.run(main)


def test_self_cancellation(kernel):
    async def suicidal_task():
        task = await current_task()
        await task.cancel(blocking=False)
        # Cancellation is delivered the next time we block
        with pytest.raises(CancelledError):
            await sleep(0)

    kernel.run(suicidal_task)

def test_disable_cancellation_explicit_raise(kernel):
    async def task():
        with pytest.raises(CancelledError):
             async with disable_cancellation():
                 raise CancelledError()

    kernel.run(task)

def test_set_cancellation(kernel):
    async def main():
        await set_cancellation(CancelledError())
        assert not await check_cancellation(TaskTimeout)
        with pytest.raises(CancelledError):
            await sleep(1)

        await set_cancellation(CancelledError())
        async with disable_cancellation():
            # This should return the given exception and clear it
            assert isinstance(await check_cancellation(CancelledError), CancelledError)
            # Verify that cleared
            assert not await check_cancellation()

    kernel.run(main())

def test_taskgroup_misc(kernel):
    async def main():
         g = TaskGroup()
         await g.cancel_remaining()

         g = TaskGroup()
         await g.spawn(sleep, 0)
         await g.cancel_remaining()

    kernel.run(main)

async def producer(ch):
    c = await ch.accept(authkey=b'peekaboo')
    for i in range(10):
        await c.send(i)
    await c.send(None)   # Sentinel

def test_contextvars():
    import contextvars
    x = contextvars.ContextVar('x', default=0)
    from curio.task import ContextTask
    events = []
    async def countdown(n):
        x.set(n)
        while x.get() > 0:
            events.append(x.get())
            x.set(x.get() - 1)
            await sleep(0)

    async def main():
        async with TaskGroup() as g:
            await g.spawn(countdown, 3)
            await g.spawn(countdown, 6)

    with Kernel(taskcls=ContextTask) as kernel:
        kernel.run(main)

    assert events == [3,6,2,5,1,4,3,2,1]

def test_task_group_result(kernel):
    async def child(x, y):
        return x + y

    async def main():
        async with TaskGroup(wait=any) as g:
            await g.spawn(child, 1, 1)
            await g.spawn(child, 2, 2)
            await g.spawn(child, 3, 3)

        assert g.result == 2

    kernel.run(main())

# Smoke test on diagnostic functions (for code coverage)
def test_task_diag(kernel):
    async def child():
        await sleep(0.25)

    async def main():
        t = await spawn(child)
        s = str(t)
        s = t.where()
        s = t.traceback()
        del t

    kernel.run(main())

def test_late_join(kernel):
    async def child():
        pass

    async def main():
        t = await spawn(child)
        await sleep(0.1)
        await t.cancel()
        assert t.joined
        assert t.terminated
        assert not t.cancelled

    kernel.run(main)


def test_task_group_join_done(kernel):
    async def add(x, y):
        return x + y

    async def task():
        async with TaskGroup(wait=all) as w:
            await w.spawn(add, 1, 1)
            await w.spawn(add, 2, 2)
            t3 = await w.spawn(add, 3, 3)
            r3 = await t3.join()
            assert r3 == 6

        assert w.results == [2, 4]

    kernel.run(task)

def test_errors(kernel):
    # Test for premature result in task  group
    async def f():
        g = TaskGroup()
        with pytest.raises(RuntimeError):
            r = g.result
        with pytest.raises(RuntimeError):
            e = g.exception
        with pytest.raises(RuntimeError):
            r = g.results
        with pytest.raises(RuntimeError):
            r = g.exceptions

        with pytest.raises(RuntimeError):
            r = await g.next_result()

        await g.join()

    kernel.run(f)

def test_diag(kernel):
    # Test ability to get stack traces from different contexts
    from curio.task import _get_stack

    # Normal generator
    def gen(n):
        yield n

    def gen2(n):
        yield from gen(n)

    # Async generator
    async def agen(n):
        yield (await sleep(0.1))

    g = gen(0)
    next(g)
    w = _get_stack(g)

    g = gen2(0)
    next(g)
    w = _get_stack(g)

    async def f():
        a = agen(0)
        w = _get_stack(a)
        await a.asend(None)

    g = f()
    g.send(None)
    w = _get_stack(g)



# test_network.py
from os.path import dirname, join
import sys
import os
import ssl
from functools import partial
import pytest

from curio import *
from curio import network
from curio import ssl as curiossl
from curio.socket import *

def test_tcp_echo(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        while True:
            results.append('recv wait')
            data = await client.recv(100)
            if not data:
                break
            results.append(('handler', data))
            await client.sendall(data)
        results.append('handler done')

    async def client(address, serv):
        results.append('client start')
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.send(b'Msg1')
        await sleep(0.1)
        resp = await sock.recv(100)
        results.append(('client', resp))
        await sock.send(b'Msg2')
        await sleep(0.1)
        resp = await sock.recv(100)
        results.append(('client', resp))
        results.append('client close')
        await sock.close()
        await serv.cancel()

    async def main():
        async with TaskGroup() as g:
            serv = await g.spawn(tcp_server, '', 25000, handler)
            await g.spawn(client, ('localhost', 25000), serv)

    kernel.run(main())

    assert results == [
        'client start',
        'handler start',
        'recv wait',
        ('handler', b'Msg1'),
        'recv wait',
        ('client', b'Msg1'),
        ('handler', b'Msg2'),
        'recv wait',
        ('client', b'Msg2'),
        'client close',
        'handler done'
    ]

if not sys.platform.startswith('win'):
    def test_unix_echo(kernel):
        results = []

        async def handler(client, addr):
            results.append('handler start')
            while True:
                results.append('recv wait')
                data = await client.recv(100)
                if not data:
                    break
                results.append(('handler', data))
                await client.sendall(data)
            results.append('handler done')

        async def client(address, serv):
            results.append('client start')
            sock = await network.open_unix_connection(address)
            await sock.send(b'Msg1')
            await sleep(0.1)
            resp = await sock.recv(100)
            results.append(('client', resp))
            await sock.send(b'Msg2')
            await sleep(0.1)
            resp = await sock.recv(100)
            results.append(('client', resp))
            results.append('client close')
            await sock.close()
            await serv.cancel()

        async def main():
            try:
                os.remove('/tmp/curionet')
            except OSError:
                pass
            async with TaskGroup() as g:
                serv = await g.spawn(unix_server, '/tmp/curionet', handler)
                await g.spawn(client, '/tmp/curionet', serv)

        kernel.run(main())

        assert results == [
            'client start',
            'handler start',
            'recv wait',
            ('handler', b'Msg1'),
            'recv wait',
            ('client', b'Msg1'),
            ('handler', b'Msg2'),
            'recv wait',
            ('client', b'Msg2'),
            'client close',
            'handler done'
        ]


def test_ssl_server(kernel):

    async def client(host, port, context):
        sock = await network.open_connection(host, port, ssl=context, server_hostname=host)
        await sock.sendall(b'Hello, world!')
        resp = await sock.recv(4096)
        return resp

    async def handler(client_sock, addr):
        data = await client_sock.recv(1000)
        assert data == b'Hello, world!'
        await client_sock.send(b'Back atcha: ' + data)

    async def main():
        # It might be desirable to move these out of the examples
        # directory, as this test are now relying on them being around
        file_path = join(dirname(dirname(__file__)), 'examples')
        cert_file = join(file_path, 'ssl_test.crt')
        key_file = join(file_path, 'ssl_test_rsa')

        server_context = curiossl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        server_context.load_cert_chain(certfile=cert_file, keyfile=key_file)

        stdlib_client_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        curio_client_context = curiossl.create_default_context(ssl.Purpose.SERVER_AUTH)

        server_task = await spawn(partial(network.tcp_server, '', 10000, handler, ssl=server_context))
        await sleep(0.1)

        for test_context in (curio_client_context, stdlib_client_context):
            test_context.check_hostname = False
            test_context.verify_mode = ssl.CERT_NONE
            resp = await client('localhost', 10000, test_context)
            assert resp == b'Back atcha: Hello, world!'

        await server_task.cancel()

    kernel.run(main())

if not sys.platform.startswith('win'):
    def test_unix_ssl_server(kernel):
        async def client(address, context):
            sock = await network.open_unix_connection(address, ssl=context)
            await sock.sendall(b'Hello, world!')
            resp = await sock.recv(4096)
            return resp

        async def handler(client_sock, addr):
            data = await client_sock.recv(1000)
            assert data == b'Hello, world!'
            await client_sock.send(b'Back atcha: ' + data)

        async def main():
            # It might be desirable to move these out of the examples
            # directory, as this test are now relying on them being around
            file_path = join(dirname(dirname(__file__)), 'examples')
            cert_file = join(file_path, 'ssl_test.crt')
            key_file = join(file_path, 'ssl_test_rsa')

            server_context = curiossl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            server_context.load_cert_chain(certfile=cert_file, keyfile=key_file)

            stdlib_client_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            curio_client_context = curiossl.create_default_context(ssl.Purpose.SERVER_AUTH)

            try:
                os.remove('/tmp/curionet')
            except OSError:
                pass
            server_task = await spawn(partial(network.unix_server, '/tmp/curionet', handler, ssl=server_context))
            await sleep(0.1)

            for test_context in (curio_client_context, stdlib_client_context):
                test_context.check_hostname = False
                test_context.verify_mode = ssl.CERT_NONE
                resp = await client('/tmp/curionet', test_context)
                assert resp == b'Back atcha: Hello, world!'

            await server_task.cancel()

        kernel.run(main())


def test_ssl_wrapping(kernel):

    async def client(host, port, context):
        sock = await network.open_connection(host, port, ssl=context, server_hostname=host)
        await sock.sendall(b'Hello, world!')
        resp = await sock.recv(4096)
        return resp

    async def handler(client_sock, addr):
        data = await client_sock.recv(1000)
        assert data == b'Hello, world!'
        await client_sock.send(b'Back atcha: ' + data)

    def server(host, port, context):
        sock = socket(AF_INET, SOCK_STREAM)
        try:
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind((host, port))
            sock.listen(5)
            return network.run_server(sock, handler, context)
        except Exception:
            sock._socket.close()
            raise

    async def main():
        # It might be desirable to move these out of the examples
        # directory, as this test are now relying on them being around
        file_path = join(dirname(dirname(__file__)), 'examples')
        cert_file = join(file_path, 'ssl_test.crt')
        key_file = join(file_path, 'ssl_test_rsa')

        server_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        server_context.load_cert_chain(certfile=cert_file, keyfile=key_file)

        stdlib_client_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        curio_client_context = curiossl.create_default_context(ssl.Purpose.SERVER_AUTH)

        server_task = await spawn(server, 'localhost', 10000, server_context)

        for test_context in (curio_client_context, stdlib_client_context):
            test_context.check_hostname = False
            test_context.verify_mode = ssl.CERT_NONE
            resp = await client('localhost', 10000, test_context)
            assert resp == b'Back atcha: Hello, world!'

        await server_task.cancel()

    kernel.run(main())

@pytest.mark.internet
def test_ssl_outgoing(kernel):
    async def main():
        c = await network.open_connection('google.com', 443, ssl=True, server_hostname='google.com')
        await c.close()

        c = await network.open_connection('google.com', 443, ssl=True)
        await c.close()

        c = await network.open_connection('google.com', 443, ssl=True, alpn_protocols=['h2'])
        await c.close()


    kernel.run(main)

def test_ssl_manual_wrapping(kernel):

    async def client(host, port, context):
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect((host, port))
        ssl_sock = await context.wrap_socket(sock, server_hostname=host)
        await ssl_sock.sendall(b'Hello, world!')
        resp = await ssl_sock.recv(4096)
        return resp

    async def handler(client_sock, addr):
        data = await client_sock.recv(1000)
        assert data == b'Hello, world!'
        await client_sock.send(b'Back atcha: ' + data)

    def server(host, port, context):
        sock = socket(AF_INET, SOCK_STREAM)
        try:
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind((host, port))
            sock.listen(5)
            return network.run_server(sock, handler, context)
        except Exception:
            sock._socket.close()
            raise

    async def main():
        # It might be desirable to move these out of the examples
        # directory, as this test are now relying on them being around
        file_path = join(dirname(dirname(__file__)), 'examples')
        cert_file = join(file_path, 'ssl_test.crt')
        key_file = join(file_path, 'ssl_test_rsa')

        server_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        server_context.load_cert_chain(certfile=cert_file, keyfile=key_file)

        curio_client_context = curiossl.create_default_context(ssl.Purpose.SERVER_AUTH)
        server_task = await spawn(server, 'localhost', 10000, server_context)

        curio_client_context.check_hostname = False
        curio_client_context.verify_mode = ssl.CERT_NONE
        resp = await client('localhost', 10000, curio_client_context)
        assert resp == b'Back atcha: Hello, world!'

        await server_task.cancel()

    kernel.run(main())

@pytest.mark.internet
def test_errors(kernel):
    async def main():
        with pytest.raises(ValueError):
            c = await network.open_connection('google.com', 443, server_hostname='google.com')
            await c.close()

        with pytest.raises(Exception):
            c = await network.open_connection('google.com', 443, ssl=True, server_hostname='yahoo.com')
            await c.close()

        with pytest.raises(ValueError):
            await network.tcp_server('localhost', 25000, None, ssl=True)


        if not sys.platform.startswith('win'):
            with pytest.raises(OSError):
                await network.tcp_server('localhost', 80, None)

            with pytest.raises(OSError):
                await network.unix_server('/tmp', None)

            with pytest.raises(ValueError):
                c = await network.open_unix_connection('/tmp/curionet', server_hostname='google.com')
                await c.close()

    kernel.run(main)




# curio_zmq.py
#
# Curio support for ZeroMQ.   Requires pyzmq.  The following test programs
# can be used to try it out:
#
#    zmq_puller.py/zmq_pusher.py    - Push/pull socket example
#    zmq_rpclient.py/zmq_rpcserv.py - RPC Server
#
'''
ZeroMQ wrapper module
---------------------

The curio_zmq module provides an async wrapper around the third party
pyzmq library for communicating via ZeroMQ.   You use it in the same way except
that certain operations are replaced by async functions.

Context(*args, **kwargs)

   An asynchronous subclass of zmq.Context. It has the same arguments
   and methods as the synchronous class.   Create ZeroMQ sockets using the
   socket() method of this class.

Sockets created by the curio_zmq.Context() class have the following
methods replaced by asynchronous versions:

   Socket.send(data, flags=0, copy=True, track=False)
   Socket.recv(flags=0, copy=True, track=False)
   Socket.send_multipart(msg_parts, flags=0, copy=True, track=False)
   Socket.recv_multipart(flags=0, copy=True, track=False)
   Socket.send_pyobj(obj, flags=0, protocol=pickle.DEFAULT_PROTOCOL)
   Socket.recv_pyobj(flags=0)
   Socket.send_json(obj, flags=0, **kwargs)
   Socket.recv_json(flags, **kwargs)
   Socket.send_string(u, flags=0, copy=True, encoding='utf-8')
   Socket.recv_string(flags=0, encoding='utf-8')

Here is an example of task that uses a ZMQ PUSH socket::

    import curio_zmq as zmq

    async def pusher(address):
        ctx = zmq.Context()
        sock = ctx.socket(zmq.PUSH)
        sock.bind(address)
        for n in range(100):
            await sock.send(b'Message %d' % n)
        await sock.send(b'exit')

    if __name__ == '__main__':
        zmq.run(pusher('tcp://*:9000'))

Here is an example of a Curio task that receives messages::

    import curio_zmq as zmq

    async def puller(address):
        ctx = zmq.Context()
        sock = ctx.socket(zmq.PULL)
        sock.connect(address)
        while True:
            msg = await sock.recv()
            if msg == b'exit':
                break
            print('Got:', msg)

    if __name__ == '__main__':
        zmq.run(puller('tcp://localhost:9000'))
'''

import pickle
from zmq.utils import jsonapi
import zmq

from curio.kernel import run  # for import compatibility
from curio.traps import _read_wait, _write_wait

# Pull all ZMQ constants and exceptions into our namespace
globals().update((key, val) for key, val in vars(zmq).items()
                 if key.isupper() or
                    (isinstance(val, type) and issubclass(val, zmq.ZMQBaseError)))

class CurioZMQSocket(zmq.Socket):

    async def send(self, data, flags=0, copy=True, track=False):
        while True:
            try:
                return super().send(data, flags | zmq.NOBLOCK, copy, track)
            except zmq.Again:
                await _write_wait(self)

    async def recv(self, flags=0, copy=True, track=False):
         while True:
             try:
                 return super().recv(flags | zmq.NOBLOCK, copy, track)
             except zmq.Again:
                 await _read_wait(self)

    async def send_multipart(self, msg_parts, flags=0, copy=True, track=False):
        for msg in msg_parts[:-1]:
            await self.send(msg, zmq.SNDMORE | flags, copy=copy, track=track)
        return await self.send(msg_parts[-1], flags, copy=copy, track=track)

    async def recv_multipart(self, flags=0, copy=True, track=False):
         parts = [ await self.recv(flags, copy=copy, track=track) ]
         while self.getsockopt(zmq.RCVMORE):
             parts.append(await self.recv(flags, copy=copy, track=track))
         return parts

    async def send_pyobj(self, obj, flags=0, protocol=pickle.DEFAULT_PROTOCOL):
        return await self.send(pickle.dumps(obj, protocol), flags)

    async def recv_pyobj(self, flags=0):
        return pickle.loads(await self.recv(flags))

    async def send_json(self, obj, flags=0, **kwargs):
        return await self.send(jsonapi.dumps(obj, **kwargs), flags)

    async def recv_json(self, flags, **kwargs):
        return jsonapi.loads(await self.recv(flags), **kwargs)

    async def send_string(self, u, flags=0, copy=True, encoding='utf-8'):
        return await self.send(u.encode(encoding), flags=flags, copy=copy)

    async def recv_string(self, flags=0, encoding='utf-8'):
        return (await self.recv(flags=flags)).decode(encoding)

class Context(zmq.Context):
    _socket_class = CurioZMQSocket
